@@ -17,20 +17,36 @@ const SPECIAL_BASE_VERTEX: &str = "base_vertex";
 const SPECIAL_BASE_INSTANCE: &str = "base_instance";
 const SPECIAL_OTHER: &str = "other";
 
+struct EpStructMember {
+    name: String,
+    ty: Handle<crate::Type>,
+    // technically, this should always be `Some`
+    binding: Option<crate::Binding>,
+    index: u32,
+}
+
 /// Structure contains information required for generating
 /// wrapped structure of all entry points arguments
-pub(super) struct EntryPointBinding {
+struct EntryPointBinding {
+    /// Name of the fake EP argument that contains the struct
+    /// with all the flattened input data.
+    arg_name: String,
     /// Generated structure name
-    name: String,
+    ty_name: String,
     /// Members of generated structure
     members: Vec<EpStructMember>,
 }
 
-struct EpStructMember {
-    name: String,
-    ty: Handle<crate::Type>,
-    binding: Option<crate::Binding>,
-    index: usize,
+pub(super) struct EntryPointInterface {
+    /// If `Some`, the input of an entry point is gathered in a special
+    /// struct with members sorted by binding.
+    /// The `EntryPointBinding::members` array is sorted by index,
+    /// so that we can walk it in `write_ep_arguments_initialization`.
+    input: Option<EntryPointBinding>,
+    /// If `Some`, the output of an entry point is flattened.
+    /// The `EntryPointBinding::members` array is sorted by binding,
+    /// So that we can walk it in `Statement::Return` handler.
+    output: Option<EntryPointBinding>,
 }
 
 #[derive(Clone, Eq, PartialEq, PartialOrd, Ord)]
@@ -50,19 +66,6 @@ impl InterfaceKey {
     }
 }
 
-// Returns true for structures that need their members permuted,
-// so that first come the user-defined varyings
-// in ascending locations, and then built-ins. This allows VS and FS
-// interfaces to match with regards to order.
-fn needs_permutation(members: &[crate::StructMember]) -> bool {
-    //Note: this is a bit of a hack. We need to re-order the output fields, but we can only do this
-    // for non-layouted structures. It may be possible for an WGSL program can use the same struct
-    // for both host sharing and the interface. This case isn't supported here.
-    let has_layout = members.iter().any(|m| m.offset != 0);
-    let has_binding = members.iter().any(|m| m.binding.is_some());
-    has_binding && !has_layout
-}
-
 #[derive(Copy, Clone, PartialEq)]
 enum Io {
     Input,
@@ -76,7 +79,7 @@ impl<'a, W: fmt::Write> super::Writer<'a, W> {
             names: crate::FastHashMap::default(),
             namer: proc::Namer::default(),
             options,
-            ep_inputs: Vec::new(),
+            entry_point_io: Vec::new(),
             named_expressions: crate::NamedExpressions::default(),
             wrapped_array_lengths: crate::FastHashSet::default(),
             wrapped_image_queries: crate::FastHashSet::default(),
@@ -88,7 +91,7 @@ impl<'a, W: fmt::Write> super::Writer<'a, W> {
         self.names.clear();
         self.namer
             .reset(module, super::keywords::RESERVED, &[], &mut self.names);
-        self.ep_inputs.clear();
+        self.entry_point_io.clear();
         self.named_expressions.clear();
         self.wrapped_array_lengths.clear();
         self.wrapped_image_queries.clear();
@@ -199,8 +202,8 @@ impl<'a, W: fmt::Write> super::Writer<'a, W> {
 
         // Write all entry points wrapped structs
         for ep in module.entry_points.iter() {
-            let ep_input = self.write_ep_input_struct(module, &ep.function, ep.stage, &ep.name)?;
-            self.ep_inputs.push(ep_input);
+            let ep_io = self.write_ep_interface(module, &ep.function, ep.stage, &ep.name)?;
+            self.entry_point_io.push(ep_io);
         }
 
         // Write all regular functions
@@ -306,6 +309,8 @@ impl<'a, W: fmt::Write> super::Writer<'a, W> {
         Ok(super::ReflectionInfo { entry_point_names })
     }
 
+    //TODO: we could force fragment outputs to always go through `entry_point_io.output` path
+    // if they are struct, so that the `stage` argument here could be omitted.
     fn write_semantic(
         &mut self,
         binding: &crate::Binding,
@@ -328,67 +333,209 @@ impl<'a, W: fmt::Write> super::Writer<'a, W> {
         Ok(())
     }
 
+    fn write_interface_struct(
+        &mut self,
+        module: &Module,
+        shader_stage: (ShaderStage, Io),
+        struct_name: String,
+        mut members: Vec<EpStructMember>,
+    ) -> Result<EntryPointBinding, Error> {
+        // Sort the members so that first come the user-defined varyings
+        // in ascending locations, and then built-ins. This allows VS and FS
+        // interfaces to match with regards to order.
+        members.sort_by_key(|m| InterfaceKey::new(m.binding.as_ref()));
+
+        write!(self.out, "struct {}", struct_name)?;
+        writeln!(self.out, " {{")?;
+        for m in members.iter() {
+            write!(self.out, "{}", back::INDENT)?;
+            self.write_type(module, m.ty)?;
+            write!(self.out, " {}", &m.name)?;
+            if let Some(ref binding) = m.binding {
+                self.write_semantic(binding, Some(shader_stage))?;
+            }
+            writeln!(self.out, ";")?;
+        }
+        writeln!(self.out, "}};")?;
+        writeln!(self.out)?;
+
+        match shader_stage.1 {
+            Io::Input => {
+                // bring back the original order
+                members.sort_by_key(|m| m.index);
+            }
+            Io::Output => {
+                // keep it sorted by binding
+            }
+        }
+
+        Ok(EntryPointBinding {
+            arg_name: self.namer.call_unique(struct_name.to_lowercase().as_str()),
+            ty_name: struct_name,
+            members,
+        })
+    }
+
+    /// Flatten all entry point arguments into a single struct.
+    /// This is needed since we need to re-order them: first placing user locations,
+    /// then built-ins.
     fn write_ep_input_struct(
         &mut self,
         module: &Module,
         func: &crate::Function,
         stage: ShaderStage,
         entry_point_name: &str,
-    ) -> Result<Option<EntryPointBinding>, Error> {
-        Ok(if !func.arguments.is_empty() {
-            let struct_name_prefix = match stage {
-                ShaderStage::Vertex => "VertexInput",
-                ShaderStage::Fragment => "FragmentInput",
-                ShaderStage::Compute => "ComputeInput",
-            };
-            let struct_name = format!("{}_{}", struct_name_prefix, entry_point_name);
+    ) -> Result<EntryPointBinding, Error> {
+        let struct_name = format!("{:?}Input_{}", stage, entry_point_name);
 
-            let mut members = Vec::with_capacity(func.arguments.len());
-            for (index, arg) in func.arguments.iter().enumerate() {
-                let member_name = if let Some(ref name) = arg.name {
-                    name
-                } else {
-                    "member"
-                };
-                members.push(EpStructMember {
-                    name: self.namer.call_unique(member_name),
-                    ty: arg.ty,
-                    binding: arg.binding.clone(),
-                    index,
-                });
-            }
-
-            // Sort the members so that first come the user-defined varyings
-            // in ascending locations, and then built-ins. This allows VS and FS
-            // interfaces to match with regards to order.
-            members.sort_by_key(|m| InterfaceKey::new(m.binding.as_ref()));
-
-            write!(self.out, "struct {}", &struct_name)?;
-            writeln!(self.out, " {{")?;
-            for m in members.iter() {
-                write!(self.out, "{}", back::INDENT)?;
-                self.write_type(module, m.ty)?;
-                write!(self.out, " {}", &m.name)?;
-                if let Some(ref binding) = m.binding {
-                    self.write_semantic(binding, Some((stage, Io::Input)))?;
+        let mut fake_members = Vec::new();
+        for arg in func.arguments.iter() {
+            match module.types[arg.ty].inner {
+                TypeInner::Struct { ref members, .. } => {
+                    for member in members.iter() {
+                        let member_name = if let Some(ref name) = member.name {
+                            name
+                        } else {
+                            "member"
+                        };
+                        let index = fake_members.len() as u32;
+                        fake_members.push(EpStructMember {
+                            name: self.namer.call_unique(member_name),
+                            ty: member.ty,
+                            binding: member.binding.clone(),
+                            index,
+                        });
+                    }
+                }
+                _ => {
+                    let member_name = if let Some(ref name) = arg.name {
+                        name
+                    } else {
+                        "member"
+                    };
+                    let index = fake_members.len() as u32;
+                    fake_members.push(EpStructMember {
+                        name: self.namer.call_unique(member_name),
+                        ty: arg.ty,
+                        binding: arg.binding.clone(),
+                        index,
+                    });
                 }
-                writeln!(self.out, ";")?;
             }
-            writeln!(self.out, "}};")?;
-            writeln!(self.out)?;
+        }
 
-            // now bring back the old order
-            members.sort_by_key(|m| m.index);
+        self.write_interface_struct(module, (stage, Io::Input), struct_name, fake_members)
+    }
 
-            Some(EntryPointBinding {
-                name: struct_name,
-                members,
-            })
-        } else {
-            None
+    /// Flatten all entry point results into a single struct.
+    /// This is needed since we need to re-order them: first placing user locations,
+    /// then built-ins.
+    fn write_ep_output_struct(
+        &mut self,
+        module: &Module,
+        result: &crate::FunctionResult,
+        stage: ShaderStage,
+        entry_point_name: &str,
+    ) -> Result<EntryPointBinding, Error> {
+        let struct_name = format!("{:?}Output_{}", stage, entry_point_name);
+
+        let mut fake_members = Vec::new();
+        let empty = [];
+        let members = match module.types[result.ty].inner {
+            TypeInner::Struct { ref members, .. } => members,
+            ref other => {
+                log::error!("Unexpected {:?} output type without a binding", other);
+                &empty[..]
+            }
+        };
+
+        for member in members.iter() {
+            let member_name = if let Some(ref name) = member.name {
+                name
+            } else {
+                "member"
+            };
+            let index = fake_members.len() as u32;
+            fake_members.push(EpStructMember {
+                name: self.namer.call_unique(member_name),
+                ty: member.ty,
+                binding: member.binding.clone(),
+                index,
+            });
+        }
+
+        self.write_interface_struct(module, (stage, Io::Output), struct_name, fake_members)
+    }
+
+    /// Writes special interface structures for an entry point. The special structures have
+    /// all the fields flattened into them and sorted by binding. They are only needed for
+    /// VS outputs and FS inputs, so that these interfaces match.
+    fn write_ep_interface(
+        &mut self,
+        module: &Module,
+        func: &crate::Function,
+        stage: ShaderStage,
+        ep_name: &str,
+    ) -> Result<EntryPointInterface, Error> {
+        Ok(EntryPointInterface {
+            input: if !func.arguments.is_empty() && stage == ShaderStage::Fragment {
+                Some(self.write_ep_input_struct(module, func, stage, ep_name)?)
+            } else {
+                None
+            },
+            output: match func.result {
+                Some(ref fr) if fr.binding.is_none() && stage == ShaderStage::Vertex => {
+                    Some(self.write_ep_output_struct(module, fr, stage, ep_name)?)
+                }
+                _ => None,
+            },
         })
     }
 
+    /// Write an entry point preface that initializes the arguments as specified in IR.
+    fn write_ep_arguments_initialization(
+        &mut self,
+        module: &Module,
+        func: &crate::Function,
+        ep_index: u16,
+    ) -> BackendResult {
+        let ep_input = match self.entry_point_io[ep_index as usize].input.take() {
+            Some(ep_input) => ep_input,
+            None => return Ok(()),
+        };
+        let mut fake_iter = ep_input.members.iter();
+        for (arg_index, arg) in func.arguments.iter().enumerate() {
+            write!(self.out, "{}", back::INDENT)?;
+            self.write_type(module, arg.ty)?;
+            let arg_name = &self.names[&NameKey::EntryPointArgument(ep_index, arg_index as u32)];
+            write!(self.out, " {}", arg_name)?;
+            match module.types[arg.ty].inner {
+                TypeInner::Array { size, .. } => {
+                    self.write_array_size(module, size)?;
+                    let fake_member = fake_iter.next().unwrap();
+                    writeln!(self.out, " = {}.{};", ep_input.arg_name, fake_member.name)?;
+                }
+                TypeInner::Struct { ref members, .. } => {
+                    write!(self.out, " = {{ ")?;
+                    for index in 0..members.len() {
+                        if index != 0 {
+                            write!(self.out, ", ")?;
+                        }
+                        let fake_member = fake_iter.next().unwrap();
+                        write!(self.out, "{}.{}", ep_input.arg_name, fake_member.name)?;
+                    }
+                    writeln!(self.out, " }};")?;
+                }
+                _ => {
+                    let fake_member = fake_iter.next().unwrap();
+                    writeln!(self.out, " = {}.{};", ep_input.arg_name, fake_member.name)?;
+                }
+            }
+        }
+        assert!(fake_iter.next().is_none());
+        Ok(())
+    }
+
     /// Helper method used to write global variables
     /// # Notes
     /// Always adds a newline
@@ -571,28 +718,18 @@ impl<'a, W: fmt::Write> super::Writer<'a, W> {
         module: &Module,
         handle: Handle<crate::Type>,
         _block: bool,
-        original_members: &[crate::StructMember],
+        members: &[crate::StructMember],
         shader_stage: Option<(ShaderStage, Io)>,
     ) -> BackendResult {
         // Write struct name
-        write!(self.out, "struct {}", self.names[&NameKey::Type(handle)])?;
-        writeln!(self.out, " {{")?;
+        let struct_name = &self.names[&NameKey::Type(handle)];
+        writeln!(self.out, "struct {} {{", struct_name)?;
 
-        //TODO: avoid heap allocation
-        let mut members = original_members
-            .iter()
-            .enumerate()
-            .map(|(index, m)| (index, m.ty, m.binding.clone()))
-            .collect::<Vec<_>>();
-        if needs_permutation(original_members) {
-            members.sort_by_key(|&(_, _, ref binding)| InterfaceKey::new(binding.as_ref()));
-        }
-
-        for (index, ty, binding) in members {
+        for (index, member) in members.iter().enumerate() {
             // The indentation is only for readability
             write!(self.out, "{}", back::INDENT)?;
 
-            match module.types[ty].inner {
+            match module.types[member.ty].inner {
                 TypeInner::Array {
                     base,
                     size,
@@ -629,7 +766,7 @@ impl<'a, W: fmt::Write> super::Writer<'a, W> {
                         interpolation,
                         sampling,
                         ..
-                    }) = binding
+                    }) = member.binding
                     {
                         if let Some(interpolation) = interpolation {
                             write!(self.out, "{} ", interpolation.to_hlsl_str())?
@@ -642,12 +779,12 @@ impl<'a, W: fmt::Write> super::Writer<'a, W> {
                         }
                     }
 
-                    if let TypeInner::Matrix { .. } = module.types[ty].inner {
+                    if let TypeInner::Matrix { .. } = module.types[member.ty].inner {
                         write!(self.out, "row_major ")?;
                     }
 
                     // Write the member type and name
-                    self.write_type(module, ty)?;
+                    self.write_type(module, member.ty)?;
                     write!(
                         self.out,
                         " {}",
@@ -656,7 +793,7 @@ impl<'a, W: fmt::Write> super::Writer<'a, W> {
                 }
             }
 
-            if let Some(ref binding) = binding {
+            if let Some(ref binding) = member.binding {
                 self.write_semantic(binding, shader_stage)?;
             };
             writeln!(self.out, ";")?;
@@ -748,7 +885,7 @@ impl<'a, W: fmt::Write> super::Writer<'a, W> {
         Ok(())
     }
 
-    /// Helper method used to write structs
+    /// Helper method used to write functions
     /// # Notes
     /// Ends in a newline
     fn write_function(
@@ -760,7 +897,18 @@ impl<'a, W: fmt::Write> super::Writer<'a, W> {
     ) -> BackendResult {
         // Function Declaration Syntax - https://docs.microsoft.com/en-us/windows/win32/direct3dhlsl/dx-graphics-hlsl-function-syntax
         if let Some(ref result) = func.result {
-            self.write_type(module, result.ty)?;
+            match func_ctx.ty {
+                back::FunctionType::Function(_) => {
+                    self.write_type(module, result.ty)?;
+                }
+                back::FunctionType::EntryPoint(index) => {
+                    if let Some(ref ep_output) = self.entry_point_io[index as usize].output {
+                        write!(self.out, "{}", ep_output.ty_name)?;
+                    } else {
+                        self.write_type(module, result.ty)?;
+                    }
+                }
+            }
         } else {
             write!(self.out, "void")?;
         }
@@ -772,6 +920,9 @@ impl<'a, W: fmt::Write> super::Writer<'a, W> {
         match func_ctx.ty {
             back::FunctionType::Function(handle) => {
                 for (index, arg) in func.arguments.iter().enumerate() {
+                    if index != 0 {
+                        write!(self.out, ", ")?;
+                    }
                     // Write argument type
                     let arg_ty = match module.types[arg.ty].inner {
                         // pointers in function arguments are expected and resolve to `inout`
@@ -789,31 +940,33 @@ impl<'a, W: fmt::Write> super::Writer<'a, W> {
 
                     // Write argument name. Space is important.
                     write!(self.out, " {}", argument_name)?;
-                    if index < func.arguments.len() - 1 {
-                        // Add a separator between args
-                        write!(self.out, ", ")?;
+                    if let TypeInner::Array { size, .. } = module.types[arg.ty].inner {
+                        self.write_array_size(module, size)?;
                     }
                 }
             }
-            back::FunctionType::EntryPoint(index) => {
-                // EntryPoint arguments wrapped into structure
-                // We need to ensure that entry points have arguments too.
-                // For the case when we working with multiple entry points
-                // for example vertex shader with arguments and fragment shader without arguments.
-                if !self.ep_inputs.is_empty()
-                    && !module.entry_points[index as usize]
-                        .function
-                        .arguments
-                        .is_empty()
-                {
-                    if let Some(ref ep_input) = self.ep_inputs[index as usize] {
-                        write!(
-                            self.out,
-                            "{} {}",
-                            ep_input.name,
-                            self.namer
-                                .call_unique(ep_input.name.to_lowercase().as_str())
-                        )?;
+            back::FunctionType::EntryPoint(ep_index) => {
+                if let Some(ref ep_input) = self.entry_point_io[ep_index as usize].input {
+                    write!(self.out, "{} {}", ep_input.ty_name, ep_input.arg_name,)?;
+                } else {
+                    let stage = module.entry_points[ep_index as usize].stage;
+                    for (index, arg) in func.arguments.iter().enumerate() {
+                        if index != 0 {
+                            write!(self.out, ", ")?;
+                        }
+                        self.write_type(module, arg.ty)?;
+
+                        let argument_name =
+                            &self.names[&NameKey::EntryPointArgument(ep_index, index as u32)];
+
+                        write!(self.out, " {}", argument_name)?;
+                        if let TypeInner::Array { size, .. } = module.types[arg.ty].inner {
+                            self.write_array_size(module, size)?;
+                        }
+
+                        if let Some(ref binding) = arg.binding {
+                            self.write_semantic(binding, Some((stage, Io::Input)))?;
+                        }
                     }
                 }
             }
@@ -822,21 +975,25 @@ impl<'a, W: fmt::Write> super::Writer<'a, W> {
         write!(self.out, ")")?;
 
         // Write semantic if it present
-        let stage = match func_ctx.ty {
-            back::FunctionType::EntryPoint(index) => {
-                Some(module.entry_points[index as usize].stage)
-            }
-            _ => None,
-        };
-        if let Some(ref result) = func.result {
-            if let Some(ref binding) = result.binding {
-                self.write_semantic(binding, stage.map(|s| (s, Io::Output)))?;
+        if let back::FunctionType::EntryPoint(index) = func_ctx.ty {
+            let stage = module.entry_points[index as usize].stage;
+            if let Some(crate::FunctionResult {
+                binding: Some(ref binding),
+                ..
+            }) = func.result
+            {
+                self.write_semantic(binding, Some((stage, Io::Output)))?;
             }
         }
 
         // Function body start
         writeln!(self.out)?;
         writeln!(self.out, "{{")?;
+
+        if let back::FunctionType::EntryPoint(index) = func_ctx.ty {
+            self.write_ep_arguments_initialization(module, func, index)?;
+        }
+
         // Write function local variables
         for (handle, local) in func.local_variables.iter() {
             // Write indentation (only for readability)
@@ -979,22 +1136,47 @@ impl<'a, W: fmt::Write> super::Writer<'a, W> {
                     // We can safery unwrap here, since we now we working with struct
                     let ty = base_ty_res.handle().unwrap();
                     let struct_name = &self.names[&NameKey::Type(ty)];
-                    let variable_name = self.namer.call_unique(struct_name.as_str()).to_lowercase();
+                    let variable_name = self.namer.call(&struct_name.to_lowercase());
                     write!(
                         self.out,
                         "{}const {} {} = ",
                         INDENT.repeat(indent),
                         struct_name,
-                        variable_name
+                        variable_name,
                     )?;
                     self.write_expr(module, expr, func_ctx)?;
                     writeln!(self.out, ";")?;
-                    writeln!(
-                        self.out,
-                        "{}return {};",
-                        INDENT.repeat(indent),
-                        variable_name
-                    )?;
+
+                    // for entry point returns, we may need to reshuffle the outputs into a different struct
+                    let ep_output = match func_ctx.ty {
+                        back::FunctionType::Function(_) => None,
+                        back::FunctionType::EntryPoint(index) => {
+                            self.entry_point_io[index as usize].output.as_ref()
+                        }
+                    };
+                    let final_name = match ep_output {
+                        Some(ep_output) => {
+                            let final_name = self.namer.call_unique(&variable_name);
+                            write!(
+                                self.out,
+                                "{}const {} {} = {{ ",
+                                INDENT.repeat(indent),
+                                ep_output.ty_name,
+                                final_name,
+                            )?;
+                            for (index, m) in ep_output.members.iter().enumerate() {
+                                if index != 0 {
+                                    write!(self.out, ", ")?;
+                                }
+                                let member_name = &self.names[&NameKey::StructMember(ty, m.index)];
+                                write!(self.out, "{}.{}", variable_name, member_name)?;
+                            }
+                            writeln!(self.out, " }};")?;
+                            final_name
+                        }
+                        None => variable_name,
+                    };
+                    writeln!(self.out, "{}return {};", INDENT.repeat(indent), final_name)?;
                 } else {
                     write!(self.out, "{}return ", INDENT.repeat(indent))?;
                     self.write_expr(module, expr, func_ctx)?;
@@ -1201,8 +1383,61 @@ impl<'a, W: fmt::Write> super::Writer<'a, W> {
                 self.temp_access_chain = chain;
                 self.named_expressions.insert(result, res_name);
             }
-            Statement::Switch { .. } => {
-                return Err(Error::Unimplemented(format!("write_stmt {:?}", stmt)))
+            Statement::Switch {
+                selector,
+                ref cases,
+                ref default,
+            } => {
+                // Start the switch
+                write!(self.out, "{}", INDENT.repeat(indent))?;
+                write!(self.out, "switch(")?;
+                self.write_expr(module, selector, func_ctx)?;
+                writeln!(self.out, ") {{")?;
+
+                // Write all cases
+                let indent_str_1 = INDENT.repeat(indent + 1);
+                let indent_str_2 = INDENT.repeat(indent + 2);
+
+                for case in cases {
+                    writeln!(self.out, "{}case {}: {{", &indent_str_1, case.value)?;
+
+                    if case.fall_through {
+                        // Generate each fallthrough case statement in a new block. This is done to
+                        // prevent symbol collision of variables declared in these cases statements.
+                        writeln!(self.out, "{}/* fallthrough */", &indent_str_2)?;
+                        writeln!(self.out, "{}{{", &indent_str_2)?;
+                    }
+                    for sta in case.body.iter() {
+                        self.write_stmt(
+                            module,
+                            sta,
+                            func_ctx,
+                            indent + 2 + usize::from(case.fall_through),
+                        )?;
+                    }
+
+                    if case.fall_through {
+                        writeln!(self.out, "{}}}", &indent_str_2)?;
+                    } else {
+                        writeln!(self.out, "{}break;", &indent_str_2)?;
+                    }
+
+                    writeln!(self.out, "{}}}", &indent_str_1)?;
+                }
+
+                // Only write the default block if the block isn't empty
+                // Writing default without a block is valid but it's more readable this way
+                if !default.is_empty() {
+                    writeln!(self.out, "{}default: {{", &indent_str_1)?;
+
+                    for sta in default {
+                        self.write_stmt(module, sta, func_ctx, indent + 2)?;
+                    }
+
+                    writeln!(self.out, "{}}}", &indent_str_1)?;
+                }
+
+                writeln!(self.out, "{}}}", INDENT.repeat(indent))?
             }
         }
 
@@ -1269,24 +1504,9 @@ impl<'a, W: fmt::Write> super::Writer<'a, W> {
         match *expression {
             Expression::Constant(constant) => self.write_constant(module, constant)?,
             Expression::Compose { ty, ref components } => {
-                let (braces_init, permutation) = match module.types[ty].inner {
-                    TypeInner::Struct { ref members, .. } => {
-                        let permutation = if needs_permutation(members) {
-                            //TODO: avoid heap allocation. We can pre-compute this at the module leve.
-                            let mut permutation = members
-                                .iter()
-                                .enumerate()
-                                .map(|(index, m)| (index, InterfaceKey::new(m.binding.as_ref())))
-                                .collect::<Vec<_>>();
-                            permutation.sort_by_key(|&(_, ref key)| key.clone());
-                            Some(permutation)
-                        } else {
-                            None
-                        };
-                        (true, permutation)
-                    }
-                    TypeInner::Array { .. } => (true, None),
-                    _ => (false, None),
+                let braces_init = match module.types[ty].inner {
+                    TypeInner::Struct { .. } | TypeInner::Array { .. } => true,
+                    _ => false,
                 };
 
                 if braces_init {
@@ -1296,16 +1516,12 @@ impl<'a, W: fmt::Write> super::Writer<'a, W> {
                     write!(self.out, "(")?;
                 }
 
-                for index in 0..components.len() {
+                for (index, &component) in components.iter().enumerate() {
                     if index != 0 {
                         // The leading space is for readability only
                         write!(self.out, ", ")?;
                     }
-                    let comp_index = match permutation {
-                        Some(ref perm) => perm[index].0,
-                        None => index,
-                    };
-                    self.write_expr(module, components[comp_index], func_ctx)?;
+                    self.write_expr(module, component, func_ctx)?;
                 }
 
                 if braces_init {
@@ -1400,24 +1616,14 @@ impl<'a, W: fmt::Write> super::Writer<'a, W> {
                 }
             }
             Expression::FunctionArgument(pos) => {
-                match func_ctx.ty {
-                    back::FunctionType::Function(handle) => {
-                        let name = &self.names[&NameKey::FunctionArgument(handle, pos)];
-                        write!(self.out, "{}", name)?;
-                    }
+                let key = match func_ctx.ty {
+                    back::FunctionType::Function(handle) => NameKey::FunctionArgument(handle, pos),
                     back::FunctionType::EntryPoint(index) => {
-                        // EntryPoint arguments wrapped into structure
-                        // We can safery unwrap here, because if we write function arguments it means, that ep_input struct already exists
-                        let ep_input = self.ep_inputs[index as usize].as_ref().unwrap();
-                        let member_name = &ep_input.members[pos as usize].name;
-                        write!(
-                            self.out,
-                            "{}.{}",
-                            &ep_input.name.to_lowercase(),
-                            member_name
-                        )?
+                        NameKey::EntryPointArgument(index, pos)
                     }
                 };
+                let name = &self.names[&key];
+                write!(self.out, "{}", name)?;
             }
             Expression::ImageSample {
                 image,
@@ -1858,9 +2064,9 @@ impl<'a, W: fmt::Write> super::Writer<'a, W> {
         components: &[Handle<crate::Constant>],
     ) -> BackendResult {
         let (open_b, close_b) = match module.types[ty].inner {
-            TypeInner::Struct { .. } => ("{ ", " }"),
+            TypeInner::Array { .. } | TypeInner::Struct { .. } => ("{ ", " }"),
             _ => {
-                // We should write type only for non struct constants
+                // We should write type only for non struct/array constants
                 self.write_type(module, ty)?;
                 ("(", ")")
             }
