@@ -350,7 +350,13 @@ impl<'a> ResolveContext<'a> {
                 })
             }
             crate::Expression::Load { pointer } => match *past(pointer).inner_with(types) {
-                Ti::Pointer { base, class: _ } => TypeResolution::Handle(base),
+                Ti::Pointer { base, class: _ } => {
+                    if let Ti::Atomic { kind, width } = types[base].inner {
+                        TypeResolution::Value(Ti::Scalar { kind, width })
+                    } else {
+                        TypeResolution::Handle(base)
+                    }
+                }
                 Ti::ValuePointer {
                     size,
                     kind,
@@ -368,7 +374,7 @@ impl<'a> ResolveContext<'a> {
             crate::Expression::ImageSample { image, .. }
             | crate::Expression::ImageLoad { image, .. } => match *past(image).inner_with(types) {
                 Ti::Image { class, .. } => TypeResolution::Value(match class {
-                    crate::ImageClass::Depth => Ti::Scalar {
+                    crate::ImageClass::Depth { multi: _ } => Ti::Scalar {
                         kind: crate::ScalarKind::Float,
                         width: 4,
                     },
@@ -377,7 +383,7 @@ impl<'a> ResolveContext<'a> {
                         width: 4,
                         size: crate::VectorSize::Quad,
                     },
-                    crate::ImageClass::Storage(format) => Ti::Vector {
+                    crate::ImageClass::Storage { format, .. } => Ti::Vector {
                         kind: format.into(),
                         width: 4,
                         size: crate::VectorSize::Quad,
@@ -395,12 +401,12 @@ impl<'a> ResolveContext<'a> {
                             kind: crate::ScalarKind::Sint,
                             width: 4,
                         },
-                        crate::ImageDimension::D2 => Ti::Vector {
+                        crate::ImageDimension::D2 | crate::ImageDimension::Cube => Ti::Vector {
                             size: crate::VectorSize::Bi,
                             kind: crate::ScalarKind::Sint,
                             width: 4,
                         },
-                        crate::ImageDimension::D3 | crate::ImageDimension::Cube => Ti::Vector {
+                        crate::ImageDimension::D3 => Ti::Vector {
                             size: crate::VectorSize::Tri,
                             kind: crate::ScalarKind::Sint,
                             width: 4,
@@ -502,6 +508,21 @@ impl<'a> ResolveContext<'a> {
                 | crate::BinaryOperator::ShiftLeft
                 | crate::BinaryOperator::ShiftRight => past(left).clone(),
             },
+            crate::Expression::AtomicResult {
+                kind,
+                width,
+                comparison,
+            } => {
+                if comparison {
+                    TypeResolution::Value(Ti::Vector {
+                        size: crate::VectorSize::Bi,
+                        kind,
+                        width,
+                    })
+                } else {
+                    TypeResolution::Value(Ti::Scalar { kind, width })
+                }
+            }
             crate::Expression::Select { accept, .. } => past(accept).clone(),
             crate::Expression::Derivative { axis: _, expr } => past(expr).clone(),
             crate::Expression::Relational { .. } => TypeResolution::Value(Ti::Scalar {
@@ -660,7 +681,7 @@ impl<'a> ResolveContext<'a> {
                     )))
                 }
             },
-            crate::Expression::Call(function) => {
+            crate::Expression::CallResult(function) => {
                 let result = self.functions[function]
                     .result
                     .as_ref()
